# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import os

from abc import ABC, abstractmethod
from typing import Final, List, Mapping, NamedTuple, Sequence

from pydub import AudioSegment

from open_dubbing.ffmpeg import FFmpeg
from open_dubbing.utterance import Utterance


class Voice(NamedTuple):
    name: str
    gender: str
    region: str = ""


class TextToSpeech(ABC):

    def __init__(self):
        self._SSML_MALE: Final[str] = "Male"
        self._SSML_FEMALE: Final[str] = "Female"
        self._DEFAULT_SPEED: Final[float] = 1.0

    @abstractmethod
    def get_available_voices(self, language_code: str) -> List[Voice]:
        pass

    def get_voices_with_region_preference(
        self, *, voices: List[Voice], target_language_region: str
    ) -> List[Voice]:
        if len(target_language_region) == 0:
            return voices

        voices_copy = voices[:]

        for voice in voices:
            if voice.region.endswith(target_language_region):
                voices_copy.remove(voice)
                voices_copy.insert(0, voice)

        return voices_copy

    def assign_voices(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        target_language: str,
        target_language_region: str,
    ) -> Mapping[str, str | None]:

        voices = self.get_available_voices(target_language)
        voices = self.get_voices_with_region_preference(
            voices=voices, target_language_region=target_language_region
        )

        voice_assignment = {}
        if len(voices) == 0:
            voice_assignment["speaker_01"] = "ona"
        else:
            for chunk in utterance_metadata:
                speaker_id = chunk["speaker_id"]
                if speaker_id in voice_assignment:
                    continue

                gender = chunk["gender"]
                for voice in voices:
                    if voice.gender.lower() == gender.lower():
                        voice_assignment[speaker_id] = voice.name
                        break

        logging.info(f"text_to_speech.assign_voices. Returns: {voice_assignment}")
        return voice_assignment

    def _convert_to_mp3(self, input_file, output_mp3):
        FFmpeg().convert_to_format(source=input_file, target=output_mp3)
        os.remove(input_file)

    def _add_text_to_speech_properties(
        self,
        *,
        utterance_metadata: Mapping[str, str | float],
    ) -> Mapping[str, str | float]:
        """Updates utterance metadata with Text-To-Speech properties."""
        utterance_metadata_copy = utterance_metadata.copy()
        voice_properties = dict(
            speed=self._DEFAULT_SPEED,
        )
        utterance_metadata_copy.update(voice_properties)
        return utterance_metadata_copy

    def update_utterance_metadata(
        self,
        *,
        utterance: Utterance | None = None,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        assigned_voices: Mapping[str, str] | None,
    ) -> Sequence[Mapping[str, str | float]]:
        """Updates utterance metadata with assigned voices."""
        updated_utterance_metadata = []
        for metadata_item in utterance_metadata:
            new_utterance = metadata_item.copy()

            fields = (
                utterance.get_modified_utterance_fields(new_utterance)
                if utterance
                else []
            )
            # If "assigned_voice" has changed we give it priority and not overwrite it
            # by recalculating from speaker_id/gender
            if not utterance or (
                "speaker_id" in fields and "assigned_voice" not in fields
            ):
                speaker_id = new_utterance.get("speaker_id")
                new_utterance["assigned_voice"] = assigned_voices.get(speaker_id)
                new_utterance = self._add_text_to_speech_properties(
                    utterance_metadata=new_utterance
                )
            updated_utterance_metadata.append(new_utterance)
        return updated_utterance_metadata

    @abstractmethod
    def get_languages(self):
        pass

    """ TTS add silence at the end that we want to remove to prevent increasing the speech of next
        segments if is not necessary."""

    def _convert_text_to_speech_without_end_silence(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        speed: float,
    ) -> str:

        dubbed_file = self._convert_text_to_speech(
            assigned_voice=assigned_voice,
            target_language=target_language,
            output_filename=output_filename,
            text=text,
            speed=speed,
        )

        dubbed_audio = AudioSegment.from_file(dubbed_file)
        pre_duration = len(dubbed_audio)

        FFmpeg().remove_silence(filename=dubbed_file)
        dubbed_audio = AudioSegment.from_file(dubbed_file)
        post_duration = len(dubbed_audio)
        if pre_duration != post_duration:
            logging.debug(
                f"text_to_speech._convert_text_to_speech_without_end_silence. File {dubbed_file} shorten from {pre_duration} to {post_duration}"
            )

        return dubbed_file

    @abstractmethod
    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        speed: float,
    ) -> str:
        pass

    def _calculate_target_utterance_speed(
        self,
        *,
        start: float,
        end: float,
        dubbed_file: str,
        utterance_metadata: Sequence[Mapping[str, float | str]],
        audio_file=str,
    ) -> float:
        """Returns the ratio between the reference and target duration."""

        end = self.get_start_time_of_next_speech_utterance(
            utterance_metadata=utterance_metadata,
            start=start,
            end=end,
            audio_file=audio_file,
        )

        reference_length = end - start
        dubbed_audio = AudioSegment.from_file(dubbed_file)
        dubbed_duration = dubbed_audio.duration_seconds
        r = (
            math.ceil(dubbed_duration / reference_length * 10) / 10
        )  # Rounds up with .1 decimal precision
        logging.debug(f"text_to_speech._calculate_target_utterance_speed: {r}")
        return r

    def _does_voice_supports_speeds(self):
        return False

    def get_start_time_of_next_speech_utterance(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        start: float,
        end: float,
        audio_file: str,
    ) -> int:
        result = None
        for utterance in utterance_metadata:
            u_start = utterance["start"]
            if u_start <= start:
                continue

            for_dubbing = utterance["for_dubbing"]
            if not for_dubbing:
                continue

            result = u_start
            break

        if not result:
            try:
                background_audio = AudioSegment.from_mp3(audio_file)
                total_duration = background_audio.duration_seconds
                logging.debug(
                    f"get_start_time_of_next_speech_utterance. File duration: {total_duration}"
                )
                result = total_duration
            except Exception as e:
                logging.error(f"Error '{e}' reading {audio_file}")

        if not result:
            result = end

        logging.debug(
            f"get_start_time_of_next_speech_utterance from_time: {start}, result: {result}"
        )
        return result

    def dub_utterances(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        output_directory: str,
        target_language: str,
        audio_file: str,
        adjust_speed: bool = True,
    ) -> Sequence[Mapping[str, str | float]]:
        """Processes a list of utterance metadata, generating dubbed audio files."""

        logging.debug(f"TextToSpeech.dub_utterances: adjust_speed: {adjust_speed}")
        updated_utterance_metadata = []
        for utterance in utterance_metadata:
            utterance_copy = utterance.copy()
            if not utterance_copy["for_dubbing"]:
                try:
                    dubbed_path = utterance_copy["path"]
                except KeyError:
                    dubbed_path = f"chunk_{utterance['start']}_{utterance['end']}.mp3"
            else:
                assigned_voice = utterance_copy["assigned_voice"]
                text = utterance_copy["translated_text"]
                try:
                    path = utterance_copy["path"]
                    base_filename = os.path.splitext(os.path.basename(path))[0]
                    output_filename = os.path.join(
                        output_directory, f"dubbed_{base_filename}.mp3"
                    )
                except KeyError:
                    output_filename = os.path.join(
                        output_directory,
                        f"dubbed_chunk_{utterance['start']}_{utterance['end']}.mp3",
                    )

                speed = utterance_copy["speed"]
                dubbed_path = self._convert_text_to_speech_without_end_silence(
                    assigned_voice=assigned_voice,
                    target_language=target_language,
                    output_filename=output_filename,
                    text=text,
                    speed=speed,
                )
                assigned_voice = utterance_copy.get("assigned_voice", None)
                assigned_voice = assigned_voice if assigned_voice else ""
                support_speeds = self._does_voice_supports_speeds()

                start = utterance["start"]
                end = utterance["end"]
                speed = self._calculate_target_utterance_speed(
                    start=start,
                    end=end,
                    dubbed_file=dubbed_path,
                    utterance_metadata=utterance_metadata,
                    audio_file=audio_file,
                )

                logging.debug(f"support_speeds: {support_speeds}, speed: {speed}")

                if speed > 1.0:
                    translated_text = utterance_copy["translated_text"]
                    logging.debug(
                        f"text_to_speech.dub_utterances. Need to increase speed for '{translated_text}'"
                    )

                    MAX_SPEED = 1.3
                    if speed > MAX_SPEED:
                        logging.debug(
                            f"text_to_speech.dub_utterances: Reduced speed from {speed} to {MAX_SPEED}"
                        )
                        speed = MAX_SPEED

                    translated_text = utterance_copy["translated_text"]
                    logging.debug(
                        f"text_to_speech.dub_utterances: Adjusting speed to {speed} for '{translated_text}'"
                    )

                    utterance_copy["speed"] = speed
                    if support_speeds:
                        dubbed_path = self._convert_text_to_speech_without_end_silence(
                            assigned_voice=assigned_voice,
                            target_language=target_language,
                            output_filename=output_filename,
                            text=text,
                            speed=speed,
                        )
                    else:
                        FFmpeg().adjust_audio_speed(
                            filename=dubbed_path,
                            speed=speed,
                        )
                        logging.debug(
                            f"text_to_speech.adjust_audio_speed: dubbed_audio: {dubbed_path}, speed: {speed}"
                        )

            utterance_copy["dubbed_path"] = dubbed_path
            updated_utterance_metadata.append(utterance_copy)
        return updated_utterance_metadata



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import re
import time

from abc import ABC, abstractmethod
from typing import Final, Mapping, Sequence

_BREAK_MARKER: Final[str] = "<BREAK>"


class Translation(ABC):

    def __init__(self, device="cpu"):
        self.device = device

    @abstractmethod
    def load_model(self):
        pass

    def _generate_script(self, *, utterance_metadata, key: str = "text") -> str:
        """Generates a script string from a list of utterance metadata."""
        trimmed_lines = [
            item[key].strip() if item[key] else "" for item in utterance_metadata
        ]
        logging.debug(f"translation.generate_script. Input: {trimmed_lines}")
        r = _BREAK_MARKER + _BREAK_MARKER.join(trimmed_lines) + _BREAK_MARKER
        logging.debug(f"translation.generate_script. Returns: {r}")
        return r

    @abstractmethod
    def get_language_pairs(self):
        pass

    @abstractmethod
    def _translate_text(
        self, source_language: str, target_language: str, text: str
    ) -> str:
        pass

    def translate_utterances(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        source_language: str,
        target_language: str,
    ) -> Sequence[Mapping[str, str | float]]:

        script = self._generate_script(utterance_metadata=utterance_metadata)
        translated_script = self._translate_script(
            script=script,
            source_language=source_language,
            target_language=target_language,
        )
        return self._add_translations(
            utterance_metadata=utterance_metadata,
            translated_script=translated_script,
        )

    def _translate_script(
        self,
        *,
        script: str,
        source_language: str,
        target_language: str,
    ) -> str:
        """Translates the provided transcript to the target language.
        Input: <BREAK>Hello, my name is Jordi Mas.<BREAK>I'm from Barcelona<BREAK>and I also work in Barcelona.<BREAK>Thanks a lot for listening.<BREAK>For.<BREAK>
        Returns <BREAK>Hello, my name is Jordi Mas.<BREAK>I'm from Barcelona<BREAK>and I also work in Barcelona.<BREAK>Thanks a lot for listening.<BREAK>For.<BREAK>

        """

        start_time = time.time()

        logging.debug(f"translation.translate_script. Input script: {script}")
        logging.debug(
            f"translation.translate_script. Input source_language: {source_language}, target_language: {target_language}"
        )

        # Split the input string by the <BREAK> delimiter
        parts = script.split(_BREAK_MARKER)

        translated_parts = []
        for text in parts:
            if len(text.strip()) == 0:
                translation = ""
            else:
                translation = self._translate_text(
                    source_language=source_language,
                    target_language=target_language,
                    text=text,
                )

            translated_parts.append(translation)

        translation = _BREAK_MARKER.join(translated_parts)
        logging.debug(f"translation.translate_script. Translation: {translation}")

        pretty_data = json.dumps(translation, indent=4, ensure_ascii=False)
        logging.debug(f"translation.translate_script. Returns: {pretty_data}")
        execution_time = time.time() - start_time
        logging.debug(
            "translation.translate_script. Time: %.2f seconds.", execution_time
        )

        return translation

    def _add_translations(
        self,
        *,
        utterance_metadata: Sequence[Mapping[str, str | float]],
        translated_script: str,
    ) -> Sequence[Mapping[str, str | float]]:
        """Updates the "translated_text" field of each utterance metadata with the corresponding text segment."""
        stripped_translation = re.sub(
            rf"^\s*{_BREAK_MARKER}\s*|\s*{_BREAK_MARKER}\s*$", "", translated_script
        )
        text_segments = stripped_translation.split(_BREAK_MARKER)
        if len(utterance_metadata) != len(text_segments):
            raise ValueError(
                "The utterance metadata must be of the same length as the text"
                f" segments. Currently they are: {len(utterance_metadata)} and"
                f" {len(text_segments)}."
            )
        updated_utterance_metadata = []
        for metadata, translated_text in zip(utterance_metadata, text_segments):
            updated_utterance_metadata.append(
                {**metadata, "translated_text": translated_text}
            )

        pretty_data = json.dumps(
            updated_utterance_metadata, indent=4, ensure_ascii=False
        )
        logging.debug(f"translation.translated_script. Input: {translated_script}")
        logging.debug(f"translation.translated_script. Returns: {pretty_data}")
        return updated_utterance_metadata



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import subprocess
import tempfile

from typing import List


class FFmpeg:

    def _run(self, *, command: List[str], fail: bool = True):
        with open(os.devnull, "wb") as devnull:
            try:
                result = subprocess.run(command, stdout=devnull, stderr=subprocess.PIPE)
                if result.returncode != 0:
                    raise subprocess.CalledProcessError(result.returncode, command)
            except subprocess.CalledProcessError as e:
                logging.error(
                    f"Error running command: {command} failed with exit code {e.returncode} and output '{result.stderr}'"
                )
                if fail:
                    raise

    def convert_to_format(self, *, source: str, target: str):
        cmd = [
            "ffmpeg",
            "-hide_banner",
            "-y",
            "-i",
            source,
            target,
        ]
        FFmpeg()._run(command=cmd)

    def remove_silence(self, *, filename: str):
        tmp_filename = ""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            tmp_filename = temp_file.name
            shutil.copyfile(filename, tmp_filename)
            cmd = [
                "ffmpeg",
                "-hide_banner",
                "-y",
                "-i",
                tmp_filename,
                "-af",
                "silenceremove=stop_periods=-1:stop_duration=0.1:stop_threshold=-50dB",
                filename,
            ]
            FFmpeg()._run(command=cmd, fail=False)

        if os.path.exists(tmp_filename):
            os.remove(tmp_filename)

    def adjust_audio_speed(self, *, filename: str, speed: float):
        tmp_filename = ""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            tmp_filename = temp_file.name
            shutil.copyfile(filename, tmp_filename)
            cmd = [
                "ffmpeg",
                "-hide_banner",
                "-y",
                "-i",
                tmp_filename,
                "-filter:a",
                f"atempo={speed}",
                filename,
            ]
            FFmpeg()._run(command=cmd, fail=False)

        if os.path.exists(tmp_filename):
            os.remove(tmp_filename)

    def embed_subtitles(
        self,
        *,
        video_file: str,
        subtitles_files: List[str],
        languages_iso_639_3: List[str],
    ):
        filename = ""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            shutil.copyfile(video_file, temp_file.name)
            output_file = video_file  # Modify in-place

            cmd = [
                "ffmpeg",
                "-y",  # Overwrite output files without asking
                "-i",
                temp_file.name,
            ]

            # Add subtitle inputs
            for subtitles_file in subtitles_files:
                cmd.extend(["-i", subtitles_file])

            # Map streams
            cmd.append("-map")
            cmd.append("0")  # Map all streams from the main video file
            for idx, language in enumerate(languages_iso_639_3):
                cmd.extend(
                    [
                        "-map",
                        str(idx + 1),  # Map each subtitle file
                        "-c:s",
                        "mov_text",  # Subtitle codec
                        "-metadata:s:s:" + str(idx),
                        f"language={language}",
                    ]
                )

            # Add codecs for video and audio
            cmd.extend(["-c:v", "copy", "-c:a", "copy", output_file])

            logging.debug(f"embed_subtitles. Command: {' '.join(cmd)}")

            # Run the command using the _run method
            self._run(command=cmd, fail=False)
            filename = temp_file.name

        if os.path.exists(filename):
            os.remove(filename)

    @staticmethod
    def is_ffmpeg_installed():
        cmd = ["ffprobe", "-version"]
        try:
            if (
                subprocess.run(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ).returncode
                == 0
            ):
                return True
        except FileNotFoundError:
            return False
        return False



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum


class ExitCode(IntEnum):
    INVALID_LANGUAGE_SPT = 100
    INVALID_LANGUAGE_TRANS = 101
    INVALID_LANGUAGE_TTS = 102
    INVALID_FILEFORMAT = 103
    MISSING_HF_KEY = 104
    NO_FFMPEG = 105
    NO_COQUI_TTS = 106
    NO_COQUI_ESPEAK = 107
    NO_CLI_CFG_FILE = 108
    NO_APERTIUM_SERVER = 109
    NO_TTS_API_SERVER = 110
    UPDATE_MISSING_FILES = 111



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import functools
import logging
import os
import re
import shutil
import sys
import time

from typing import Final

import psutil
import torch

from pyannote.audio import Pipeline

from open_dubbing import audio_processing
from open_dubbing.demucs import Demucs
from open_dubbing.exit_code import ExitCode
from open_dubbing.ffmpeg import FFmpeg
from open_dubbing.preprocessing import PreprocessingArtifacts
from open_dubbing.speech_to_text import SpeechToText
from open_dubbing.subtitles import Subtitles
from open_dubbing.text_to_speech import TextToSpeech
from open_dubbing.translation import Translation
from open_dubbing.utterance import Utterance
from open_dubbing.video_processing import VideoProcessing

_DEFAULT_PYANNOTE_MODEL: Final[str] = "pyannote/speaker-diarization-3.1"
_NUMBER_OF_STEPS: Final[int] = 7


@dataclasses.dataclass
class PostprocessingArtifacts:
    """Instance with postprocessing outputs.

    Attributes:
        audio_file: A path to a dubbed audio file.
        video_file: A path to a dubbed video file. The video is optional.
    """

    audio_file: str
    video_file: str | None


class PyAnnoteAccessError(Exception):
    """Error when establishing access to PyAnnore from Hugging Face."""

    pass


def rename_input_file(original_input_file: str) -> str:
    """Converts a filename to lowercase letters and numbers only, preserving the file extension.

    Args:
        original_filename: The filename to normalize.

    Returns:
        The normalized filename.
    """
    directory, filename = os.path.split(original_input_file)
    base_name, extension = os.path.splitext(filename)
    normalized_name = re.sub(r"[^a-z0-9]", "", base_name.lower())
    return os.path.join(directory, normalized_name + extension)


def overwrite_input_file(input_file: str, updated_input_file: str) -> None:
    """Renames a file in place to lowercase letters and numbers only, preserving the file extension."""

    if not os.path.exists(input_file):
        raise FileNotFoundError(f"File '{input_file}' not found.")

    shutil.move(input_file, updated_input_file)


class Dubber:
    """A class to manage the entire ad dubbing process."""

    def __init__(
        self,
        *,
        input_file: str,
        output_directory: str,
        source_language: str,
        target_language: str,
        target_language_region: str,
        hugging_face_token: str | None = None,
        tts: TextToSpeech,
        translation: Translation,
        stt: SpeechToText,
        device: str,
        cpu_threads: int = 0,
        clean_intermediate_files: bool = False,
        original_subtitles: bool = False,
        dubbed_subtitles: bool = False,
    ) -> None:
        self._input_file = input_file
        self.output_directory = output_directory
        self.source_language = source_language
        self.target_language = target_language
        self.target_language_region = target_language_region
        self.pyannote_model = _DEFAULT_PYANNOTE_MODEL
        self.hugging_face_token = hugging_face_token
        self.utterance_metadata = None
        self._number_of_steps = _NUMBER_OF_STEPS
        self.tts = tts
        self.translation = translation
        self.stt = stt
        self.device = device
        self.cpu_threads = cpu_threads
        self.clean_intermediate_files = clean_intermediate_files
        self.preprocessing_output = None
        self.original_subtitles = original_subtitles
        self.dubbed_subtitles = dubbed_subtitles

        if cpu_threads > 0:
            torch.set_num_threads(cpu_threads)

    @functools.cached_property
    def input_file(self):
        renamed_input_file = rename_input_file(self._input_file)
        if renamed_input_file != self._input_file:
            logging.warning(
                "The input file was renamed because the original name contained"
                " spaces, hyphens, or other incompatible characters. The updated"
                f" input file is: {renamed_input_file}"
            )
            overwrite_input_file(
                input_file=self._input_file, updated_input_file=renamed_input_file
            )
        return renamed_input_file

    def log_maxrss_memory(self):
        if sys.platform == "win32" or sys.platform == "win64":
            return

        import resource

        max_rss_self = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
        if sys.platform == "darwin":
            return max_rss_self / 1024

        logging.info(f"Maximum memory used: {max_rss_self:.0f} MB")

    def log_debug_task_and_getime(self, text, start_time):
        process = psutil.Process(os.getpid())
        current_rss = process.memory_info().rss / 1024**2
        _time = time.time() - start_time
        logging.info(
            f"Completed task '{text}': current_rss {current_rss:.2f} MB, time {_time:.2f}s"
        )
        return _time

    @functools.cached_property
    def pyannote_pipeline(self) -> Pipeline:
        """Loads the PyAnnote diarization pipeline."""
        return Pipeline.from_pretrained(
            self.pyannote_model, use_auth_token=self.hugging_face_token
        )

    def _verify_api_access(self) -> None:
        """Verifies access to all the required APIs."""
        logging.debug("Verifying access to PyAnnote from HuggingFace.")
        if not self.pyannote_pipeline:
            raise PyAnnoteAccessError(
                "No access to HuggingFace. Make sure you passed the correct API token"
                " either as 'hugging_face_token' or through the"
                " environmental variable. Also, please make sure you accepted the"
                " user agreement for the segmentation model"
                " (https://huggingface.co/pyannote/segmentation-3.0) and the speaker"
                " diarization model"
                " (https://huggingface.co/pyannote/speaker-diarization-3.1)."
            )
        logging.debug("Access to PyAnnote from HuggingFace verified.")

    def run_preprocessing(self) -> None:
        """Splits audio/video, applies DEMUCS, and segments audio into utterances with PyAnnote."""
        video_file, audio_file = VideoProcessing.split_audio_video(
            video_file=self.input_file, output_directory=self.output_directory
        )
        demucs = Demucs()
        demucs_command = demucs.build_demucs_command(
            audio_file=audio_file,
            output_directory=self.output_directory,
            device=self.device,
        )
        demucs.execute_demucs_command(command=demucs_command)
        audio_vocals_file, audio_background_file = (
            demucs.assemble_split_audio_file_paths(command=demucs_command)
        )

        utterance_metadata = audio_processing.create_pyannote_timestamps(
            audio_file=audio_file,
            pipeline=self.pyannote_pipeline,
            device=self.device,
        )
        utterance_metadata = audio_processing.run_cut_and_save_audio(
            utterance_metadata=utterance_metadata,
            audio_file=audio_file,
            output_directory=self.output_directory,
        )
        self.utterance_metadata = utterance_metadata
        self.preprocessing_output = PreprocessingArtifacts(
            video_file=video_file,
            audio_file=audio_file,
            audio_vocals_file=audio_vocals_file,
            audio_background_file=audio_background_file,
        )

    def run_speech_to_text(self) -> None:
        """Transcribes audio, applies speaker diarization, and updates metadata with Gemini.

        Returns:
            Updated utterance metadata with speaker information and transcriptions.
        """

        media_file = (
            self.preprocessing_output.video_file
            if self.preprocessing_output.video_file
            else self.preprocessing_output.audio_file
        )
        utterance_metadata = self.stt.transcribe_audio_chunks(
            utterance_metadata=self.utterance_metadata,
            source_language=self.source_language,
            no_dubbing_phrases=[],
        )
        speaker_info = self.stt.predict_gender(
            file=media_file,
            utterance_metadata=utterance_metadata,
        )
        self.utterance_metadata = self.stt.add_speaker_info(
            utterance_metadata=utterance_metadata, speaker_info=speaker_info
        )

        utterance = Utterance(self.target_language, self.output_directory)
        self.utterance_metadata = utterance.get_without_empty_blocks(
            self.utterance_metadata
        )

    def run_translation(self) -> None:
        """Translates transcribed text and potentially merges utterances"""

        self.utterance_metadata = self.translation.translate_utterances(
            utterance_metadata=self.utterance_metadata,
            source_language=self.source_language,
            target_language=self.target_language,
        )

    def run_configure_text_to_speech(self) -> None:
        """Configures the Text-To-Speech process.

        Returns:
            Updated utterance metadata with assigned voices
            and Text-To-Speech settings.
        """
        assigned_voices = self.tts.assign_voices(
            utterance_metadata=self.utterance_metadata,
            target_language=self.target_language,
            target_language_region=self.target_language_region,
        )
        self.utterance_metadata = self.tts.update_utterance_metadata(
            utterance_metadata=self.utterance_metadata,
            assigned_voices=assigned_voices,
        )

    def run_text_to_speech(self) -> None:
        """Converts translated text to speech and dubs utterance"""
        self.utterance_metadata = self.tts.dub_utterances(
            utterance_metadata=self.utterance_metadata,
            output_directory=self.output_directory,
            target_language=self.target_language,
            audio_file=self.preprocessing_output.audio_file,
            adjust_speed=True,
        )

    def run_cleaning(self) -> None:
        if not self.clean_intermediate_files:
            return

        output_directory = None
        paths, dubbed_paths = Utterance(
            self.target_language, self.output_directory
        ).get_files_paths(self.utterance_metadata)
        for path in paths + dubbed_paths:
            if os.path.exists(path):
                os.remove(path)
            if not output_directory:
                output_directory = os.path.dirname(path)

        if output_directory:
            for path in [
                f"dubbed_audio_{self.target_language}.mp3",
                "dubbed_vocals.mp3",
            ]:
                full_path = os.path.join(output_directory, path)
                if os.path.exists(full_path):
                    os.remove(full_path)

    def run_postprocessing(self) -> None:
        """Merges dubbed audio with the original background audio and video (if applicable).

        Returns:
            Path to the final dubbed output file.
        """
        dubbed_audio_vocals_file = audio_processing.insert_audio_at_timestamps(
            utterance_metadata=self.utterance_metadata,
            background_audio_file=self.preprocessing_output.audio_background_file,
            output_directory=self.output_directory,
        )
        dubbed_audio_file = audio_processing.merge_background_and_vocals(
            background_audio_file=self.preprocessing_output.audio_background_file,
            dubbed_vocals_audio_file=dubbed_audio_vocals_file,
            output_directory=self.output_directory,
            target_language=self.target_language,
            vocals_volume_adjustment=5.0,
            background_volume_adjustment=0.0,
        )
        if not self.preprocessing_output.video_file:
            raise ValueError(
                "A video file must be provided if the input file is a video."
            )
        dubbed_video_file = VideoProcessing.combine_audio_video(
            video_file=self.preprocessing_output.video_file,
            dubbed_audio_file=dubbed_audio_file,
            output_directory=self.output_directory,
            target_language=self.target_language,
        )
        self.postprocessing_output = PostprocessingArtifacts(
            audio_file=dubbed_audio_file,
            video_file=dubbed_video_file,
        )

    def _save_utterances(self):
        metadata = {
            "source_language": self.source_language,
            "original_subtitles": self.original_subtitles,
            "dubbed_subtitles": self.dubbed_subtitles,
        }
        Utterance(self.target_language, self.output_directory).save_utterances(
            utterance_metadata=self.utterance_metadata,
            preprocessing_output=self.preprocessing_output,
            metadata=metadata,
        )

    def update(self):
        times = {}
        start_time = time.time()
        task_start_time = time.time()

        logging.info("Update dubbing process started")

        try:
            utterance = Utterance(self.target_language, self.output_directory)
            self.utterance_metadata, self.preprocessing_output, _ = (
                utterance.load_utterances()
            )
        except Exception as e:
            logging.error(
                f"Unable to read metadata at '{self.output_directory}. "
                f"Cannot find a previous execution to update. Error: '{e}'"
            )
            exit(ExitCode.UPDATE_MISSING_FILES)

        _, dubbed_paths = utterance.get_files_paths(self.utterance_metadata)
        for path in dubbed_paths:
            if not os.path.exists(path):
                logging.error(
                    f"Cannot do update operation since file '{path}' is missing."
                )
                exit(ExitCode.UPDATE_MISSING_FILES)

        # Update voices in case they have changed
        modified_utterances = utterance.get_modified_utterances(self.utterance_metadata)

        assigned_voices = self.tts.assign_voices(
            utterance_metadata=modified_utterances,
            target_language=self.target_language,
            target_language_region=self.target_language_region,
        )

        modified_utterances = self.tts.update_utterance_metadata(
            utterance=utterance,
            utterance_metadata=modified_utterances,
            assigned_voices=assigned_voices,
        )

        self.tts.dub_utterances(
            utterance_metadata=modified_utterances,
            output_directory=self.output_directory,
            target_language=self.target_language,
            audio_file=self.preprocessing_output.audio_file,
            adjust_speed=True,
        )
        times["tts"] = self.log_debug_task_and_getime(
            "Text to speech completed", task_start_time
        )

        task_start_time = time.time()
        self.run_postprocessing()
        self.run_generate_subtitles()
        self._save_utterances()
        times["postprocessing"] = self.log_debug_task_and_getime(
            "Post processing completed", task_start_time
        )
        logging.info("Dubbing process finished.")
        total_time = time.time() - start_time
        logging.info(f"Total execution time: {total_time:.2f} secs")
        for task in times:
            _time = times[task]
            per = _time * 100 / total_time
            logging.info(f" Task '{task}' in {_time:.2f} secs ({per:.2f}%)")

        self.log_maxrss_memory()
        logging.info("Output files saved in: %s.", self.output_directory)

    def run_generate_subtitles(self):
        if not self.original_subtitles and not self.dubbed_subtitles:
            return

        subtitles = Subtitles()

        filename = f"{self.source_language}.srt"
        source_srt = subtitles.write(
            utterance_metadata=self.utterance_metadata,
            directory=self.output_directory,
            filename=filename,
            translated=False,
        )

        filename = f"{self.target_language}.srt"
        target_srt = subtitles.write(
            utterance_metadata=self.utterance_metadata,
            directory=self.output_directory,
            filename=filename,
            translated=True,
        )

        subtitles_files = []
        languages_iso_639_3 = []

        if self.original_subtitles:
            subtitles_files.append(source_srt)
            languages_iso_639_3.append(self.source_language)

        if self.dubbed_subtitles:
            subtitles_files.append(target_srt)
            languages_iso_639_3.append(self.target_language)

        FFmpeg().embed_subtitles(
            video_file=self.postprocessing_output.video_file,
            subtitles_files=subtitles_files,
            languages_iso_639_3=languages_iso_639_3,
        )
        logging.info(f"Generated subtitles for languages {languages_iso_639_3}")

    def dub(self) -> PostprocessingArtifacts:
        """Orchestrates the entire dubbing process."""
        self._verify_api_access()
        logging.info("Dubbing process starting...")
        times = {}
        start_time = time.time()

        task_start_time = time.time()
        self.run_preprocessing()
        times["preprocessing"] = self.log_debug_task_and_getime(
            "Preprocessing completed", task_start_time
        )
        logging.info("Speech to text...")
        task_start_time = time.time()
        self.run_speech_to_text()
        times["stt"] = self.log_debug_task_and_getime(
            "Speech to text completed", task_start_time
        )
        task_start_time = time.time()

        self.run_translation()
        times["translation"] = self.log_debug_task_and_getime(
            "Translation completed", task_start_time
        )

        task_start_time = time.time()
        self.run_configure_text_to_speech()
        self.run_text_to_speech()
        times["tts"] = self.log_debug_task_and_getime(
            "Text to speech completed", task_start_time
        )

        task_start_time = time.time()

        self.run_postprocessing()
        self.run_generate_subtitles()
        self._save_utterances()
        self.run_cleaning()
        times["postprocessing"] = self.log_debug_task_and_getime(
            "Post processing completed", task_start_time
        )
        logging.info("Dubbing process finished.")
        total_time = time.time() - start_time
        logging.info(f"Total execution time: {total_time:.2f} secs")
        for task in times:
            _time = times[task]
            per = _time * 100 / total_time
            logging.info(f" Task '{task}' in {_time:.2f} secs ({per:.2f}%)")

        self.log_maxrss_memory()
        logging.info("Output files saved in: %s.", self.output_directory)
        return self.postprocessing_output



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

WHISPER_MODEL_NAMES = [
    "medium",
    "large-v2",
    "large-v3",
]


class NewlinePreservingHelpFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        # Split the text by explicit newlines first
        lines = text.splitlines()
        # Then apply the default behavior for line wrapping
        wrapped_lines = []
        for line in lines:
            wrapped_lines.extend(argparse.HelpFormatter._split_lines(self, line, width))
        return wrapped_lines


class CommandLine:

    @staticmethod
    def read_parameters():
        """Parses command-line arguments and runs the dubbing process."""
        parser = argparse.ArgumentParser(
            description="AI dubbing system which uses machine learning models to automatically translate and synchronize audio dialogue into different languages",
            formatter_class=NewlinePreservingHelpFormatter,
        )
        parser.add_argument(
            "--input_file",
            required=True,
            help="Path to the input video file.",
        )
        parser.add_argument(
            "--output_directory",
            default="output/",
            help="Directory to save output files.",
        )
        parser.add_argument(
            "--source_language",
            help="Source language (ISO 639-3)",
        )
        parser.add_argument(
            "--target_language",
            required=True,
            help="Target language for dubbing (ISO 639-3).",
        )
        parser.add_argument(
            "--hugging_face_token",
            default=None,
            help="Hugging Face API token.",
        )
        parser.add_argument(
            "--tts",
            type=str,
            default="mms",
            choices=["mms", "coqui", "edge", "cli", "api"],
            help=(
                "Text to Speech engine to use. Choices are:\n"
                "'mms': Meta Multilingual Speech engine, supports +1100 languages.\n"
                "'coqui': Coqui TTS, an open-source alternative for high-quality TTS.\n"
                "'edge': Microsoft Edge TSS.\n"
                "'cli': User defined TTS invoked from command line.\n"
                "'api': Implements a user defined TTS API contract to enable non supported TTS.\n"
            ),
        )
        parser.add_argument(
            "--stt",
            type=str,
            default="auto",
            choices=["auto", "faster-whisper", "transformers"],
            help=(
                "Speech to text. Choices are:\n"
                "'auto': Autoselect best implementation.\n"
                "'faster-whisper': Faster-whisper's OpenAI whisper implementation.\n"
                "'transformers': Transformers OpenAI whisper implementation.\n"
            ),
        )
        parser.add_argument(
            "--vad",
            action="store_true",
            help="Enable VAD filter when using faster-whisper (reduces hallucinations).",
        )

        parser.add_argument(
            "--translator",
            type=str,
            default="nllb",
            choices=["nllb", "apertium"],
            help=(
                "Text to Speech engine to use. Choices are:\n"
                "'nllb': Meta's no Language Left Behind (NLLB).\n"
                "'apertium': Apertium compatible API server.\n"
            ),
        )
        parser.add_argument(
            "--apertium_server",
            type=str,
            default="",
            help=("Apertium's URL server to use"),
        )

        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            choices=["cpu", "cuda"],
            help=("Device to use"),
        )
        parser.add_argument(
            "--cpu_threads",
            type=int,
            default=0,
            help="number of threads used for CPU inference (if is not specified uses defaults for each framework)",
        )
        parser.add_argument(
            "--clean-intermediate-files",
            action="store_true",
            help="clean intermediate files used during the dubbing process",
        )

        parser.add_argument(
            "--nllb_model",
            type=str,
            default="nllb-200-3.3B",
            choices=["nllb-200-1.3B", "nllb-200-3.3B"],
            help="Meta NLLB translation model size. Choices are:\n"
            "'nllb-200-3.3B': gives best translation quality.\n"
            "'nllb-200-1.3B': is the fastest.\n",
        )

        parser.add_argument(
            "--whisper_model",
            default="large-v3",
            choices=WHISPER_MODEL_NAMES,
            help="name of the OpenAI Whisper speech to text model size to use",
        )

        parser.add_argument(
            "--target_language_region",
            default="",
            help="For some TTS you can specify the region of the language. For example, 'ES' will indicate accent from Spain.",
        )

        parser.add_argument(
            "--tts_cli_cfg_file",
            default="",
            help="JSon configuration file when using a TTS which is invoked from the command line.",
        )

        parser.add_argument(
            "--log_level",
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="Set the logging level",
        )
        parser.add_argument(
            "--tts_api_server",
            type=str,
            default="",
            help=("TTS api server URL when using the 'API' tts"),
        )
        parser.add_argument(
            "--update",
            action="store_true",
            help="Update the dubbed video produced by a previous execution with the latest changes in utterance_metadata file",
        )

        parser.add_argument(
            "--original_subtitles",
            action="store_true",
            default=False,
            help="Add original subtitles as stream in the output video",
        )
        parser.add_argument(
            "--dubbed_subtitles",
            default=False,
            action="store_true",
            help="Add dubbed subtitles as stream in the output video",
        )

        return parser.parse_args()



import dataclasses


@dataclasses.dataclass
class PreprocessingArtifacts:
    """Instance with preprocessing outputs.

    Attributes:
        video_file: A path to a video ad with no audio.
        audio_file: A path to an audio track from the ad.
        audio_vocals_file: A path to an audio track with vocals only.
        audio_background_file: A path to and audio track from the ad with removed
          vocals.
    """

    video_file: str | None
    audio_file: str
    audio_vocals_file: str | None = None
    audio_background_file: str | None = None



import os
import pathlib
from setuptools import setup, find_packages
import ast

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()


def read_version(fname="open_dubbing/__init__.py"):
    with open(fname, encoding="utf-8") as f:
        for line in f:
            if line.startswith("__version__"):
                return ast.literal_eval(line.split("=")[1].strip())

    raise ValueError("__version__ not found in the file")

def parse_requirements(fname="requirements.txt"):
    with open(HERE / fname, encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]


setup(
    name="open-dubbing",
    version=read_version(),
    description="AI dubbing system which uses machine learning models to automatically translate and synchronize audio dialogue into different languages.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Softcatala/open-dubbing",
    author="Jordi Mas",
    author_email="jmas@softcatala.org",
    license="Apache Software License 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=parse_requirements(),
    extras_require={
        "dev": ["flake8==7.*", "black==24.*", "pytest==8.*", "isort==5.13"],
        "coqui": ["coqui-tts >= 0.25.1"],
    },
    entry_points={
        "console_scripts": [
            "open-dubbing=open_dubbing.main:main",
        ]
    },
    python_requires='>=3.10'
)



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import time
import urllib

from open_dubbing.translation import Translation


class TranslationApertium(Translation):

    def load_model(self):
        pass

    def set_server(self, server):
        if not server.endswith("/"):
            server = server + "/"

        self.server = server

    def _do_api_call(self, url):
        max_retries = 3
        for attempt in range(1, max_retries + 1):
            try:
                response = urllib.request.urlopen(url)
                r = response.read().decode("utf-8")
                data = json.loads(r)
                return data["responseData"]
            except Exception:
                if attempt == max_retries:
                    logging.error(
                        "translation_apertium. Max retries reached. Could not complete translation API call."
                    )
                    raise
                else:
                    logging.warning(
                        f"translation_apertium. Could not complete translation API call, retrying attempt {attempt}."
                    )
                    time.sleep(30)

    def _translate_text(
        self, source_language: str, target_language: str, text: str
    ) -> str:

        langpair = f"{source_language}|{target_language}"
        text = urllib.parse.quote_plus(text.encode("utf-8"))
        method = "translate"
        url = f"{self.server}{method}?q={text}&langpair={langpair}&markUnknown=no"
        translated = self._do_api_call(url)
        translated = translated["translatedText"]
        return translated.rstrip()

    def get_language_pairs(self):
        method = "listPairs"
        url = f"{self.server}{method}"

        pairs = self._do_api_call(url)
        results = set()
        for pair in pairs:
            source = pair["sourceLanguage"]
            target = pair["targetLanguage"]
            if len(source) != 3 or len(target) != 3:
                logging.warning(
                    f"Discarding Apertium language pair: '{source}-{target}'"
                )
                continue

            pair = (pair["sourceLanguage"], pair["targetLanguage"])
            results.add(pair)

        return results



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from datetime import timedelta


class Subtitles:

    def write(self, *, utterance_metadata, directory, filename, translated):
        srt_file_path = os.path.join(directory, filename)

        with open(srt_file_path, "w", encoding="utf-8") as subtitles_file:
            for i, utterance in enumerate(utterance_metadata):
                start_time = self.format_srt_time(utterance["start"])
                end_time = self.format_srt_time(utterance["end"])

                idx = i + 1
                srt_content = f"{idx}\n"
                srt_content += f"{start_time} --> {end_time}\n"

                text = utterance["translated_text"] if translated else utterance["text"]
                srt_content += f"{text}\n\n"
                subtitles_file.write(srt_content)
        return srt_file_path

    @staticmethod
    def format_srt_time(seconds):
        """Format seconds as HH:MM:SS,mmm for SRT files."""
        time_delta = timedelta(seconds=seconds)
        total_seconds = int(time_delta.total_seconds())
        milliseconds = int((time_delta.total_seconds() - total_seconds) * 1000)
        hours = total_seconds // 3600
        minutes = (total_seconds % 3600) // 60
        seconds = total_seconds % 60
        return f"{hours:02}:{minutes:02}:{seconds:02},{milliseconds:03}"



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import hashlib
import json
import logging
import os
import shutil
import tempfile

from typing import Any, Dict, Final, List, Tuple

from open_dubbing.preprocessing import PreprocessingArtifacts


class Utterance:

    def __init__(self, target_language: str, output_directory: str):
        self.target_language = target_language
        self.output_directory = output_directory

    def _get_file_name(self):
        _UTTERNACE_METADATA_FILE_NAME: Final[str] = "utterance_metadata"

        target_language_suffix = "_" + self.target_language.replace("-", "_").lower()
        utterance_metadata_file = os.path.join(
            self.output_directory,
            _UTTERNACE_METADATA_FILE_NAME + target_language_suffix + ".json",
        )
        return utterance_metadata_file

    def load_utterances(self) -> tuple[Any, PreprocessingArtifacts, Any]:
        utterance_metadata_file = self._get_file_name()

        with open(utterance_metadata_file, "r") as file:
            data = json.load(file)
            utterances = data["utterances"]
            preprocessing_output = PreprocessingArtifacts(
                **data["PreprocessingArtifacts"]
            )
            metadata = data["metadata"]

        return utterances, preprocessing_output, metadata

    def save_utterances(
        self,
        *,
        utterance_metadata: str,
        preprocessing_output: str,
        metadata: Dict[str, str],
        do_hash: bool = True,
        unique_id: bool = True,
    ) -> None:
        """Saves a Python dictionary to a JSON file.

        Returns:
          A path to the saved uttterance metadata.
        """
        utterance_metadata_file = self._get_file_name()

        try:
            all_data = {}
            if unique_id:
                utterance_metadata = self._add_unique_ids(utterance_metadata)

            if do_hash:
                utterance_metadata = self._hash_utterances(utterance_metadata)

            all_data["utterances"] = utterance_metadata
            if preprocessing_output:
                all_data["PreprocessingArtifacts"] = dataclasses.asdict(
                    preprocessing_output
                )
            all_data["metadata"] = metadata

            json_data = json.dumps(all_data, ensure_ascii=False, indent=4)
            with tempfile.NamedTemporaryFile(
                mode="w", delete=False, encoding="utf-8"
            ) as temporary_file:

                temporary_file.write(json_data)
                temporary_file.flush()
                os.fsync(temporary_file.fileno())
            shutil.copy(temporary_file.name, utterance_metadata_file)
            os.remove(temporary_file.name)
            logging.debug(
                "Utterance metadata saved successfully to"
                f" '{utterance_metadata_file}'"
            )
        except Exception as e:
            logging.warning(f"Error saving utterance metadata: {e}")

    def _get_utterance_fields_to_hash(self, utterance):
        filtered_fields = {
            key: value for key, value in utterance.items() if not key.startswith("_")
        }
        return filtered_fields

    def _hash_utterances(self, utterance_metadata):
        for utterance in utterance_metadata:
            filtered_fields = self._get_utterance_fields_to_hash(utterance)
            dict_str = json.dumps(filtered_fields, sort_keys=True)
            _hash = hashlib.sha256(dict_str.encode()).hexdigest()
            utterance["_hash"] = _hash

            for field in ["assigned_voice", "speaker_id"]:
                value = utterance.get(field)
                if value:
                    utterance[f"_{field}_hash"] = hashlib.sha256(
                        value.encode()
                    ).hexdigest()

        return utterance_metadata

    def _add_unique_ids(self, utterance_metadata):
        for idx, utterance in enumerate(utterance_metadata, start=1):
            new_utterance = {"id": idx}
            new_utterance.update(utterance)
            utterance_metadata[idx - 1] = new_utterance

        return utterance_metadata

    def get_files_paths(self, utterance_metadata) -> Tuple[List[str], List[str]]:
        dubbed_paths = []
        paths = []
        for chunk in utterance_metadata:
            if "path" in chunk:
                paths.append(chunk["path"])
            if "dubbed_path" in chunk:
                dubbed_paths.append(chunk["dubbed_path"])

        return paths, dubbed_paths

    def get_modified_utterance_fields(self, utterance):
        modified = []
        for field in utterance:
            field_hash = utterance.get(f"_{field}_hash")
            if not field_hash:
                continue

            field_value = utterance[field]
            current_hash = hashlib.sha256(field_value.encode()).hexdigest()

            if current_hash != field_hash:
                modified.append(field)

        return modified

    def get_modified_utterances(self, utterance_metadata):
        modified = []
        for utterance in utterance_metadata:
            _hash_utterance = utterance["_hash"]
            filtered_fields = self._get_utterance_fields_to_hash(utterance)

            dict_str = json.dumps(filtered_fields, sort_keys=True)
            _hash = hashlib.sha256(dict_str.encode()).hexdigest()
            if _hash_utterance != _hash:
                modified.append(utterance)

        logging.info(f"Modified {len(modified)} utterances")
        return modified

    def get_without_empty_blocks(self, utterance_metadata):
        new_utterance = []

        for utterance in utterance_metadata:
            text = utterance["text"]
            if len(text) == 0:
                logging.debug(f"Removing empty block: {utterance}")
                continue

            new_utterance.append(utterance)

        return new_utterance

    def update_utterances(self, utterance_master, utterance_update):
        id_to_update = {}
        utterance_new = []

        for utterance in utterance_update:
            id = utterance["id"]
            id_to_update[id] = utterance

        for utterance in utterance_master:
            id = utterance["id"]
            update = id_to_update.get(id, None)
            if not update:
                utterance_new.append(utterance)
                continue

            operation = update.get("operation", None)
            if not operation:
                raise ValueError("No operation field defined")

            if operation == "delete":
                continue

            if operation != "update":
                raise ValueError(f"Invalid operation {operation}")

            updateable_fields = [
                "speaker_id",
                "translated_text",
                "speed",
                "assigned_voice",
                "for_dubbing",
                "gender",
                "start",
                "end",
            ]
            for field in updateable_fields:
                value = update.get(field, None)
                if not value:
                    continue

                utterance[field] = value

            utterance_new.append(utterance)

        return utterance_new



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import subprocess
import warnings

from TTS.api import TTS


class Coqui:
    """Builds a list models available per each language"""

    def __init__(self, device="cpu"):
        logging.getLogger("TTS.utils").setLevel(logging.ERROR)
        logging.getLogger("TTS.utils.audio.processor").setLevel(logging.ERROR)
        warnings.filterwarnings("ignore", category=FutureWarning)
        language_models = self._build_list_language_model()
        self.language_model = self._select_model_per_language(language_models)
        self.device = device

    @property
    def languages_model(self):
        return self.language_model

    def get_languages(self):
        return self.language_model.keys()

    """ Select first vits, then glow and then the rest"""

    def _select_model_per_language(self, language_models):
        language_model = {}
        for language in language_models:
            models = language_models[language]
            found = False
            if len(models) == 1:
                language_model[language] = models[0]
                found = True
            else:
                for model in models:
                    if "vits" in model:
                        language_model[language] = model
                        found = True
                        break

                for model in models:
                    if "glow" in model:
                        language_model[language] = model
                        found = True
                        break

            if found:
                continue

            language_model[language] = models[0]

        return language_model

    def debug_list_all_voices(self):
        for model in TTS.list_models():
            voices = TTS(model).speakers
            print(f"Model {model}: {voices}")

    def _build_list_language_model(self):
        language_models = {}

        for model in TTS.list_models():
            match = re.search(r"/([a-z]{2})/", model)
            if match:
                language = match.group(1)
                models = language_models.get(language, [])
                models.append(model)
                language_models[language] = models

        return language_models

    def get_voices_language(self, language):
        model = self.language_model[language]
        voices = TTS(model).speakers
        #      print(f"voices: {voices}")
        return voices

    def synthesize_speech(
        self, input_text, language, file_path, voice=None, audio_config=None
    ):
        model = self.language_model[language]
        tts = TTS(model).to(self.device)
        tts.tts_to_file(
            text=input_text, speaker=voice, split_sentences=False, file_path=file_path
        )

    @staticmethod
    def is_espeak_ng_installed():
        for cmd in [["espeak-ng", "--version"], ["espeak", "--version"]]:
            try:
                if (
                    subprocess.run(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    ).returncode
                    == 0
                ):
                    return True
            except FileNotFoundError:
                continue
        return False
